#!/usr/bin/perl

# Copyright (C) 2000, HELM Team.
# 
# This file is part of HELM, an Hypertextual, Electronic
# Library of Mathematics, developed at the Computer Science
# Department, University of Bologna, Italy.
# 
# HELM is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# HELM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with HELM; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# 
# For details, see the HELM World-Wide-Web page,
# http://cs.unibo.it/helm/.

# First of all, let's load HELM configuration
use Env;
my $HELM_LIB_DIR = $ENV{"HELM_LIB_DIR"};
# this should be the only fixed constant
my $DEFAULT_HELM_LIB_DIR = "/usr/local/lib/helm";
if (defined ($HELM_LIB_DIR)) {
   $HELM_LIB_PATH = $HELM_LIB_DIR."/configuration.pl";
} else {
   $HELM_LIB_PATH = $DEFAULT_HELM_LIB_DIR."/configuration.pl";
}

# <ZACK>: TODO temporary, move this setting to configuration file
# set the cache mode, may be "gzipped" or "normal"
my $cachemode = $ENV{'HTTP_GETTER_CACHE_MODE'} || 'gzipped';
if (($cachemode ne 'gzipped') and ($cachemode ne 'normal')) {
	die "Invalid HTTP_GETTER_CACHE_MODE environment variable, must be 'normal' or 'gzipped'\n";
}
# </ZACK>

# next require defines: $helm_dir, $html_link, $dtd_dir, $uris_dbm
require $HELM_LIB_PATH;

use HTTP::Daemon;
use HTTP::Status;
use HTTP::Request;
use LWP::UserAgent;
use DB_File;
use Compress::Zlib;

#CSC: mancano i controlli sulle condizioni di errore di molte funzioni
#CSC: ==> non e' robusto
#CSC: altra roba da sistemare segnata con CSC

my $d = new HTTP::Daemon LocalPort => 8081;
tie(%map, 'DB_File', $uris_dbm.".db", O_RDONLY, 0664);
print "Please contact me at: <URL:", $d->url, ">\n";
print "helm_dir: $helm_dir\n";
print "dtd_dir: $dtd_dir\n";
print "urls_of_uris.db: $uris_dbm.db\n";
print "cache mode: $cachemode\n";

$SIG{CHLD} = "IGNORE"; # do not accumulate defunct processes
$SIG{USR1} = \&update; # sent by the child to make the parent update
while (my $c = $d->accept) {
 if (fork() == 0) {
    while (my $r = $c->get_request) {
        #CSC: mancano i controlli di sicurezza
        
        my $inputuri = $r->url; 
        $inputuri =~ s/^[^?]*\?uri=(.*)/$1/;
        print "\nRequest: ".$r->url."\n\n";
        my $http_method = $r->method;
        my $http_path = $r->url->path;

        if ($http_method eq 'GET' and $http_path eq "/getciconly") {
            # finds the uri, url and filename
            my $cicuri = $inputuri;

            my $cicfilename = $cicuri;
            $cicfilename =~ s/cic:(.*)/$1/;
            $cicfilename =~ s/theory:(.*)/$1/;
#            $cicfilename = $helm_dir.$cicfilename.".xml";
# <gzip>
            my $cicurl   = $map{$cicuri};
			my $extension;
			if ($cicurl =~ /\.xml$/) {	# non gzipped file
					$extension = ".xml";
			} elsif ($cicurl =~ /\.xml\.gz$/) {	# gzipped file
					$extension = ".xml.gz";
			} else {	# error: unknown extension
				die "unexpected extension in url: $cicurl, might be '.xml' or '.xml.gz'";
			}
            $cicfilename = $helm_dir.$cicfilename.$extension;

            #my $cicurl   = $map{$cicuri};
# </gzip>
            if (!defined($cicurl)) {
             print "\nNOT FOUND!!!!!\n";
             $c->send_error(RC_NOT_FOUND)
            } else {
               print_request("cic",$cicuri,$cicurl,$cicfilename);

               # Retrieves the file
               my $ciccontent = download(0,"cic",$cicurl,$cicfilename);

               # Answering the client
               answer($c,$ciccontent);
            }
        } elsif ($http_method eq 'GET' and $http_path eq "/get") {
            # finds the uris, urls and filenames
            my $cicuri = $inputuri,
               $typesuri = $inputuri,
               $annuri = $inputuri;
            my $annsuffix;
            if ($inputuri =~ /\.types$/) {
               $cicuri    =~ s/(.*)\.types$/$1/;
               undef($annuri);
            } elsif ($inputuri =~ /\.types\.ann$/) {
               $cicuri    =~ s/(.*)\.types\.ann$/$1/;
               $typesuri  =~ s/(.*)\.ann$/$1/;
               $annsuffix = ".types.ann";
            } elsif ($inputuri =~ /\.ann$/) {
               $cicuri  =~ s/(.*)\.ann$/$1/;
               undef($typesuri);
               $annsuffix = ".ann";
            } else {
               undef($typesuri);
               undef($annuri);
            }

            my $cicfilename = $cicuri;
            $cicfilename =~ s/cic:(.*)/$1/;
            $cicfilename =~ s/theory:(.*)/$1/;
            $cicfilename = $helm_dir.$cicfilename;

#            my $typesfilename = $cicfilename.".types.xml"     if $typesuri;
#            my $annfilename  = $cicfilename.$annsuffix.".xml" if $annuri;
#            $cicfilename .= ".xml";

# <gzip>
            my $cicurl   = $map{$cicuri};
            my $typesurl = $map{$typesuri} if (defined($typesuri));
            my $annurl   = $map{$annuri}  if (defined($annuri));
			my ($cicext, $typesext, $annext);
			if ($cicurl =~ /\.xml$/) {	# normal file
				$cicext = ".xml";
			} elsif ($cicurl =~ /\.xml\.gz$/) {	# gzipped file
				$cicext = ".xml.gz";
			} else {
				die "unexpected extension in url: $cicurl; might be '.xml' or '.xml.gz'";
			}
			if (defined($typesuri)) {	# extension selection for types file
				if ($typesurl =~ /\.xml$/) {	# normal file
					$typesext = ".types.xml";
				} elsif ($typesurl =~ /\.xml\.gz$/) {	# gzipped file
					$typesext = ".types.xml.gz";
				} else {
					die "unexpected extension in url: $typesurl; might be '.xml' or '.xml.gz'";
				}
			}
			if (defined($annuri)) {	# extension selection for annotation file
				if ($annurl =~ /\.xml$/) {	# normal file
					$annext = ".xml";
				} elsif ($annurl =~ /\.xml\.gz$/) {	# gzipped file
					$annext = ".xml.gz";
				} else {
					die "unexpected extension in url: $annurl might be '.xml' or '.xml.gz'";
				}
			}
            my $typesfilename = $cicfilename.$typesext if $typesuri;
            my $annfilename  = $cicfilename.$annsuffix.$annext if $annuri;
            $cicfilename .= $cicext;
# </gzip>


            if (!defined($cicurl) ||
               (!defined($typesurl) && $typesuri) ||
               (!defined($annuri) && $annuri))
            {
             print "\nNOT FOUND!!!!!\n";
             $c->send_error(RC_NOT_FOUND)
            } else {
               print_request("cic",$cicuri,$cicurl,$cicfilename);
               print_request("types",$typesuri,$typesurl,$typesfilename)
                if ($typesuri);
               print_request("ann",$annuri,$annurl,$annfilename)
                if ($annuri);
 
               # Retrieves the files

               my $ciccontent = download(1,"cic",$cicurl,$cicfilename);
               my $typescontent =
                download(1,"types",$typesurl,$typesfilename) if ($typesuri);
               my $anncontent =
                download(1,"ann",$annurl,$annfilename) if ($annuri);
 
               # Merging the files together
 
               my $merged = <<EOT;
<?xml version="1.0" encoding="UTF-8"?>
<cicxml uri="$cicuri">
$ciccontent
$typescontent
$anncontent
</cicxml>
EOT

               # Answering the client
               answer($c,$merged);
            }
         } elsif ($http_method eq 'GET' and $http_path eq "/getdtd") {
            my $filename = $inputuri;
            $filename = $dtd_dir."/".$filename;
            print "DTD: $inputuri ==> ($filename)\n";
            if (stat($filename)) {
               print "Using local copy\n";
               open(FD, $filename) or die "Cannot open $filename\n";
               $cont = "";
               while(<FD>) { $cont .= $_; }
               close(FD);
               answer($c,$cont);
            } else {
               die "Could not find DTD!";
            }
        } elsif ($http_method eq 'GET' and $http_path eq "/conf") {
            my $quoted_html_link = $html_link;
            $quoted_html_link =~ s/&/&amp;/g;
            $quoted_html_link =~ s/</&lt;/g;
            $quoted_html_link =~ s/>/&gt;/g;
            $quoted_html_link =~ s/'/&apos;/g;
            $quoted_html_link =~ s/"/&quot;/g;
            print "\nConfiguration requested, returned #$quoted_html_link#\n";
	    $cont = "<?xml version=\"1.0\"?><html_link>$quoted_html_link</html_link>";
            answer($c,$cont);
        } elsif ($http_method eq 'GET' and $http_path eq "/update") {
           print "Update requested...";
           update();
           kill(USR1,getppid());
           print " done\n";
           answer($c,"<html><body><h1>Update done</h1></body></html>");
        } else {
            print "\nINVALID REQUEST!!!!!\n";
            $c->send_error(RC_FORBIDDEN)
        }
        print "\nRequest solved: ".$r->url."\n\n";
    }
    $c->close;
    undef($c);
    print "\nCONNECTION CLOSED\n\n";
    exit;
  } # fork
}

#================================


#CSC: Too much powerful: creates even /home, /home/users/, ...
#CSC: Does not raise errors if could not create dirs/files
sub mkdirs
{
 my ($pathname) = @_;
 my @dirs = split /\//,$pathname;
 my $tmp;
 foreach $dir (@dirs) {
  $tmp = ((defined($tmp)) ?  $tmp."\/".$dir : "");
  mkdir($tmp,0777);
 }
 rmdir($tmp);
}

sub print_request
{
 my ($str,$uri,$url,$filename) = @_;
 print $str."uri: $uri\n";
 print $str."url: $url\n";
 print $str."filename: $filename\n\n";
}

sub callback
{
 my ($data) = @_;
 $cont .= $data;
}

sub gunzip { # gunzip a file and return the deflated content
	my ($filename) = @_;

	my ($gz, $buffer, $cont);

	print "deflating $filename ...\n";
	$gz = gzopen($filename, "r") or die "Cannot open gzip'ed file $filename: $gzerrno";
	$cont = "";
	while ( $gz->gzread($buffer) > 0 ) {
		$cont .= $buffer;
	}
	die "Error while reading : $gzerrno\n" if $gzerrno != Z_STREAM_END ;
	$gz->gzclose();

	return $cont;
}

sub gzip {	# gzip the content argument and save it to filename argument
	my ($cont, $filename) = @_;

	my ($gz, $cont);

	$gz = gzopen($filename, "w") or die "Cannot gzopen for writing file $filename: $gzerrno";
	$gz->gzwrite($cont) or die "error writing: $gzerrno\n" ;
	$gz->gzclose();
}

sub download
{
 my ($remove_headers,$str,$url,$filename) = @_;
# <gzip>
 my ($gz, $buffer);

 my $resourcetype;	# retrieve mode: "normal" (.xml) or "gzipped" (.xml.gz)
 if ($filename =~ /\.xml$/) {	# set retrieve mode
	 $resourcetype = "normal";
 } elsif ($filename =~ /\.xml\.gz$/) {
	 $resourcetype = "gzipped";
 } else {
	 die "Unsupported download extension, might be '.gz' or '.xml'\n";
 }
 my $basefname = $filename;
 $basefname =~ s/\.gz$//;		# get base resource name removing trailing .gz
# </gzip>
 $cont = ""; # modified by side-effect by the callback function

 my $localfname="";
 if (stat($basefname)) {
 	$localfname=$basefname;
 } elsif (stat($basefname.".gz")) {
 	$localfname=$basefname.".gz";
 }
 if ($localfname ne "") {		# we already have local copy of requested file
								# check both possible cache entry: gzipped or normal
    print "Using local copy for the $str file\n";
# <gzip>
	if ($localfname =~ /\.xml\.gz$/) {	# deflating cached file and return it
		$cont = gunzip($localfname);
	} elsif ($localfname =~ /\.xml$/) {	# just return cached file
		open(FD, $localfname) or die "Cannot open $localfname";
		while(<FD>) { $cont .= $_; }
		close(FD);
	} else {	# error
		die "Internal error: unexpected file name $localfname, must end with '.gz' or '.xml.gz'\n";
	}
# </gzip>
 } else {	# download file from net
    print "Downloading the $str file\n";	# download file
    $ua = LWP::UserAgent->new;
    $request = HTTP::Request->new(GET => "$url");
    $response = $ua->request($request, \&callback);
               
				# cache retrieved file to disk
# <ZACK/> TODO: inefficent, I haven't yet undestood how to deflate in memory gzipped file,
#				without call "gzopen"
# <gzip>
	print "Storing the $str file\n";
	mkdirs($filename);
	open(FD, ">".$filename.".tmp") or die "Cannot open $filename.tmp\n";
	print FD $cont;
	close(FD);

	# handle cache conversion normal->gzipped or gzipped->normal as user choice
	if (($cachemode eq 'normal') and ($resourcetype eq 'normal')) {	# cache the file as is
		rename "$filename.tmp", $filename;	
	} elsif (($cachemode eq 'gzipped') and ($resourcetype eq 'gzipped')) {	# cache the file as is
											# and update the $cont variabile with deflated content
		rename "$filename.tmp", $filename;	
		$cont = gunzip($filename);
	} elsif (($cachemode eq 'normal') and ($resourcetype eq 'gzipped')) {	# deflate cache entry
																			# and update $cont
		open(FD, "> $basefname") or die "cannot open $basefname\n";
		$cont = gunzip($filename.".tmp");
		print FD $cont;
		close(FD);
		unlink "$filename.tmp";	# delete old gzipped file
	} elsif (($cachemode eq 'gzipped') and ($resourcetype eq 'normal')) {	# compress cache entry
		gzip($cont, $basefname.".gz");
		unlink "$filename.tmp";	# delete old uncompressed file
	} else {
		die "Internal error, unsopported cachemode, resourcetype couple\n";
	}
	# $cont now contained uncompressed data

# </gzip>
 }
 if ($remove_headers) {
    $cont =~ s/<\?xml [^?]*\?>//sg;
    $cont =~ s/<!DOCTYPE [^>]*>//sg;
 }
 return $cont;
}

sub answer
{
 my ($c,$cont) = @_;
 my $res = new HTTP::Response;
 $res->content($cont);
 $c->send_response($res);
}

sub update {
 untie %map;
 tie(%map, 'DB_File', $uris_dbm.".db", O_RDONLY, 0664);
}
